IMPLEMENTATION MODULE MetaStore ;


FROM SYSTEM IMPORT SIZE ;
FROM Debug IMPORT Halt ;


CONST
   Debugging = TRUE ;

TYPE
   HEAP = POINTER TO RECORD
                        UsedList,
                        FreeList    : FRAGMENT ;
                        TotalSize   : CARDINAL ;       (* used for debugging purposes *)
                        ChooseMethod: AllocateMethod ;
                     END ;

   FRAGMENT = POINTER TO RECORD
                            Left, Right: FRAGMENT ;
                            Size       : CARDINAL ;
                         END ;


(*
   InitHeap - creates a heap and returns a handle onto this heap.
              The start and size of the heap is defined by, a, and, size.
              AllocP is the allocation method chosen.
*)

PROCEDURE InitHeap (a: ADDRESS; size: CARDINAL; AllocP: AllocateMethod) : HEAP ;
VAR
   h: HEAP ;
BEGIN
   h := HEAP(a) ;
   INC(a, SIZE(h^)) ;
   WITH h^ DO
      UsedList     := NIL ;
      FreeList     := InitFragment(a, size-SIZE(h^)) ;
      TotalSize    := size-SIZE(h^) ;
      ChooseMethod := AllocP
   END ;
   IF Debugging
   THEN
      VerifyHeap(h)
   END ;
   RETURN( h )
END InitHeap ;


(*
   InitFragment - creates a new fragment from address, a, and, size.
*)

PROCEDURE InitFragment (a: ADDRESS; size: CARDINAL) : FRAGMENT ;
VAR
   f: FRAGMENT ;
BEGIN
   f := a ;
   WITH f^ DO
      Left  := f ;
      Right := f ;
      Size  := size
   END ;
   RETURN( f )
END InitFragment ;


(*
   BestFit - chooses the best fit fragment in a heap.
*)

PROCEDURE BestFit (h: HEAP; size: CARDINAL) : ADDRESS ;
VAR
   f, best : FRAGMENT ;
   fragsize: CARDINAL ;
BEGIN
   f    := h^.FreeList ;
   best := NIL ;
   IF f#NIL
   THEN
      best := FirstFit(h, size) ;
      IF best#NIL
      THEN
         fragsize := best^.Size ;
         REPEAT
            (* must check for fragment to be exactly the size we want or large enough to be fragmented *)
            IF ((f^.Size=size) OR (size<=f^.Size-SIZE(h^))) AND (f^.Size<fragsize)
            THEN
               best := f ;
               fragsize := f^.Size
            END ;
            f := f^.Right
         UNTIL f=h^.FreeList
      END
   END ;
   RETURN( best )
END BestFit ;


(*
   FirstFit - chooses the first fit fragment in a heap.
*)

PROCEDURE FirstFit (h: HEAP; size: CARDINAL) : ADDRESS ;
VAR
   f: FRAGMENT ;
BEGIN
   f := h^.FreeList ;
   IF f#NIL
   THEN
      REPEAT
         (* must check for fragment to be exactly the size we want or large enough to be fragmented *)
         IF (f^.Size=size) OR (size<=f^.Size-SIZE(h^))
         THEN
            RETURN( f )
         END ;
         f := f^.Right
      UNTIL f=h^.FreeList
   END ;
   RETURN( NIL )
END FirstFit ;


(*
   WorstFit - nah, nobody wants this do they?
*)

(* PROCEDURE WorstFit (h: HEAP; size: CARDINAL) : ADDRESS ; *)


(*
   AddToList - adds a fragment the end of a list.
*)
 
PROCEDURE AddToList (VAR h: FRAGMENT; f: FRAGMENT) ;
BEGIN
   IF h=NIL
   THEN
      h := f ;
      f^.Left := f ;
      f^.Right := f
   ELSE
      f^.Right := h ;
      f^.Left := h^.Left ;
      h^.Left^.Right := f ;
      h^.Left := f
   END
END AddToList ;
 
 
(*
   SubFromList - removes a fragment from a list.
*)
 
PROCEDURE SubFromList (VAR h: FRAGMENT; f: FRAGMENT) ;
BEGIN
   IF (f^.Left=f) AND (f=h)
   THEN
      h := NIL
   ELSE
      IF h=f
      THEN
         h := h^.Right
      END ;
      f^.Left^.Right := f^.Right ;
      f^.Right^.Left := f^.Left
   END
END SubFromList ;


(*
   Allocate - allocates, n, bytes of store from the heap.
*)

PROCEDURE Allocate (h: HEAP; n: CARDINAL) : ADDRESS ;
VAR
   f: FRAGMENT ;
BEGIN
   INC(n, SIZE(f^)) ;   (* enough space to maintain the used fragment header *)
   WITH h^ DO
      IF Debugging
      THEN
         VerifyHeap(h)
      END ;
      f := ChooseMethod(h, n) ;
      IF f=NIL
      THEN
         RETURN( NIL )
      ELSE
         SubFromList(FreeList, f) ;
         WITH f^ DO
            IF Size#n
            THEN
               AddToList(FreeList, InitFragment(ADDRESS(f)+ADDRESS(n), Size-n))
            END ;
            Size := n ;
            AddToList(UsedList, f)
         END
      END ;
      IF Debugging
      THEN
         VerifyHeap(h)
      END ;
      RETURN( f+FRAGMENT(SIZE(f^)) )
   END
END Allocate ;


(*
   Deallocate - deallocates, n, bytes of store.
*)

PROCEDURE Deallocate (h: HEAP; a: ADDRESS; size: CARDINAL) ;
VAR
   f: FRAGMENT ;
BEGIN
   IF a#NIL
   THEN
      WITH h^ DO
         f := a-ADDRESS(SIZE(f^)) ;
         IF IsOnList(UsedList, f)
         THEN
            IF Debugging
            THEN
               VerifyHeap(h)
            END ;
            SubFromList(UsedList, f) ;
            IF f^.Size=size+SIZE(f^)
            THEN
               IF Debugging
               THEN
                  IF SumList(FreeList, TotalSize)+SumList(UsedList, TotalSize)+f^.Size#h^.TotalSize
                  THEN
                     Halt('corrupt heap', __LINE__, __FILE__)
                  END
               END ;
               OrderedAddToList(FreeList, f) ;
               IF Debugging
               THEN
                  VerifyHeap(h)
               END
            ELSE
               Halt('memory leak, different size allocated to deallocated', __LINE__, __FILE__)
            END
         ELSE
            Halt('dynamic memory error, memory was never allocated', __LINE__, __FILE__)
         END
      END
   END
END Deallocate ;


(*
   Reallocate - attempts to reallocate one chunk as two. It assumes that the
                address, a, has already been allocated with fromsize bytes.
                Upon return there are two fragments allocated:

                a with tosize bytes
                b with bsize  bytes

                Note that tosize + bsize may not be equal to fromsize
                     as there is a system overhead.
*)

PROCEDURE Reallocate (h: HEAP;
                      a: ADDRESS; fromsize, tosize: CARDINAL;
                      VAR b: ADDRESS; VAR bsize: CARDINAL) ;
VAR
   f: FRAGMENT ;
BEGIN
   WITH h^ DO
      VerifyHeap(h) ;
      f := a-ADDRESS(SIZE(f^)) ;
      IF IsOnList(UsedList, f)
      THEN
         IF f^.Size=fromsize+SIZE(f^)
         THEN
            IF (tosize<SIZE(f^)) OR (tosize>f^.Size-SIZE(f^))
            THEN
               (* unable to fragment too small or too large a fragment *)
               bsize := 0 ;
               b := NIL
            ELSE
               (* looks ok *)
               SubFromList(UsedList, f) ;
               bsize := fromsize-tosize-SIZE(f^) ;   (* space for new header of second fragment *)
               (* shrink first fragment *)
               f := InitFragment(f, tosize+SIZE(f^)) ;
               (* add it to the used list again *)
               OrderedAddToList(UsedList, f) ;
               (* create second fragment with its own header *)
               b := InitFragment(f+FRAGMENT(f^.Size), bsize+SIZE(f^)) ;
               (* add it to the used list as well *)
               OrderedAddToList(UsedList, b) ;
               INC(b, ADDRESS(SIZE(f^))) ;
               VerifyHeap(h)
            END
         ELSE
            Halt('memory leak, different size allocated to deallocated', __LINE__, __FILE__)
         END
      ELSE
         Halt('dynamic memory error, memory was never allocated', __LINE__, __FILE__)
      END
   END
END Reallocate ;


(*
   OrderedAddToList - adds fragment, f, onto list, h, in the correct place.
                      It attempts to defragment the list.
*)

PROCEDURE OrderedAddToList (VAR h: FRAGMENT; f: FRAGMENT) ;
VAR
   p, q: FRAGMENT ;
BEGIN
   IF h=NIL
   THEN
      h := InitFragment(f, f^.Size)
   ELSE
      IF f<h
      THEN
         (* must add f to the beginning of the list *)
         InsertBefore(h, h, f)
      ELSIF f>h^.Left
      THEN
         (* f needs to be added to the end of the list *)
         AddToList(h, f)
      ELSE
         (* f needs to be inserted in the middle somewhere *)
         p := h ;
         q := h^.Right ;
         WHILE q<f DO
            p := q ;
            q := q^.Right
         END ;
         (* p < f < q *)
         IF (p<f) AND (f<q)
         THEN
            InsertBefore(h, q, f)
         ELSE
            Halt('list assertion failed', __LINE__, __FILE__)
         END
      END ;
      (* GrowFragments(f^.Left, f, f^.Right) *)
   END
END OrderedAddToList ;


(*
   InsertBefore - insert a fragment, new, on a circular event queue BEFORE
                  fragment, pos.
*)

PROCEDURE InsertBefore (VAR Head: FRAGMENT; pos, new: FRAGMENT) ;
BEGIN
   IF Head=NIL
   THEN
      (* empty queue *)
      Head := new ;
      new^.Right := new ;
      new^.Left := new
   ELSIF Head=pos
   THEN
      (* insert before the first element on the queue *)
      new^.Right := pos ;
      new^.Left := pos^.Left ;
      pos^.Left^.Right := new ;
      pos^.Left := new ;
      Head := new
   ELSE
      (* insert before any other element *)
      new^.Right := pos ;
      new^.Left := pos^.Left ;
      pos^.Left^.Right := new ;
      pos^.Left := new
   END
END InsertBefore ;


(*
   IsOnList - returns TRUE if, f, is on list, h.
*)

PROCEDURE IsOnList (h, f: FRAGMENT) : BOOLEAN ;
VAR
   a: FRAGMENT ;
BEGIN
   IF h#NIL
   THEN
      a := h ;
      REPEAT
         IF a=f
         THEN
            RETURN( TRUE )
         ELSE
            a := a^.Right
         END
      UNTIL a=h
   END ;
   RETURN( FALSE )
END IsOnList ;


(*
   GrowFragments - attempt to merge two contigeous fragments.
*)

PROCEDURE GrowFragments (p, r, q: FRAGMENT) ;
BEGIN
   IF p+FRAGMENT(p^.Size)=r
   THEN
      r := CombineFragments(p, r)
   END ;
   IF r+FRAGMENT(r^.Size)=q
   THEN
      r := CombineFragments(r, q)
   END
END GrowFragments ;


(*
   CombineFragments - combines two fragments, a, and, b.
*)

PROCEDURE CombineFragments (a, b: FRAGMENT) : FRAGMENT ;
BEGIN
   a^.Right := b^.Right ;
   b^.Right^.Left := a ;
   INC(a^.Size, b^.Size) ;
   RETURN( a )
END CombineFragments ;


(*
   VerifyHeap - verifies that HEAP, h, is consistant. It checks that
                the sum of the freelist of usedlist equals the total heap size.
*)

PROCEDURE VerifyHeap (h: HEAP) ;
BEGIN
   WITH h^ DO
      IF TotalSize#SumList(FreeList, TotalSize)+SumList(UsedList, TotalSize)
      THEN
         Halt('inconsistant heap, freelist and usedlist do not equal the total size', __LINE__, __FILE__)
      END
   END
END VerifyHeap ;


(*
   SumList - sums the size of FRAGMENTs on a list.
*)

PROCEDURE SumList (f: FRAGMENT; limit: CARDINAL) : CARDINAL ;
VAR
   sr, sl: CARDINAL ;
   g     : FRAGMENT ;
BEGIN
   sr := 0 ;
   IF f#NIL
   THEN
      g := f ;
      REPEAT
         INC(sr, g^.Size) ;
         IF sr>limit
         THEN
            Halt('heap is badly corrupted', __LINE__, __FILE__)
         END ;
         g := g^.Right
      UNTIL g=f ;
      sl := 0 ;
      g := f ;
      REPEAT
         INC(sl, g^.Size) ;
         IF sl>limit
         THEN
            Halt('heap is badly corrupted', __LINE__, __FILE__)
         END ;
         g := g^.Left
      UNTIL g=f ;
      IF sr#sl
      THEN
         Halt('left and right pointers give different size', __LINE__, __FILE__)
      END
   END ;
   RETURN( sr )
END SumList ;


(*
   GetLargestFree - returns the largest free space fragment size in bytes
                    in a heap. NOTE it does not allocate this fragment.
*)

PROCEDURE GetLargestFree (h: HEAP) : CARDINAL ;
VAR
   s: CARDINAL ;
   f: FRAGMENT ;
BEGIN
   s := 0 ;
   f := h^.FreeList ;
   IF f#NIL
   THEN
      REPEAT
         IF f^.Size>s
         THEN
            s := f^.Size
         END ;
         f := f^.Right
      UNTIL f=h^.FreeList
   END ;
   IF s<=SIZE(f^)
   THEN
      RETURN( 0 )
   ELSE
      RETURN( s-SIZE(f^) )    (* enough space to maintain the used fragment header *)
   END
END GetLargestFree ;


(*
   GetTotalSize - returns the total heap size.
*)

PROCEDURE GetTotalSize (h: HEAP) : CARDINAL ;
BEGIN
   IF h=NIL
   THEN
      RETURN( 0 )
   ELSE
      RETURN( h^.TotalSize )
   END
END GetTotalSize ;


(*
   GetUsedSize - returns the amount of the heap that is used.
*)

PROCEDURE GetUsedSize (h: HEAP) : CARDINAL ;
BEGIN
   WITH h^ DO
      RETURN( SumList(UsedList, TotalSize) )
   END
END GetUsedSize ;


(*
   GetFreeSize - returns the amount of the heap that is free.
*)

PROCEDURE GetFreeSize (h: HEAP) : CARDINAL ;
BEGIN
   WITH h^ DO
      RETURN( SumList(FreeList, TotalSize) )
   END
END GetFreeSize ;


(*
   GetNoOfFreeFragments - returns the number of fragments held on the free list within the heap.
*)

PROCEDURE GetNoOfFreeFragments (h: HEAP) : CARDINAL ;
BEGIN
   RETURN( CountFragments(h^.FreeList) )
END GetNoOfFreeFragments ;


(*
   GetNoOfUsedFragments - returns the number of fragments held on the used free within the heap.
*)

PROCEDURE GetNoOfUsedFragments (h: HEAP) : CARDINAL ;
BEGIN
   RETURN( CountFragments(h^.UsedList) )
END GetNoOfUsedFragments ;


(*
   CountFragments - returns the number of fragments on a list, h.
*)

PROCEDURE CountFragments (h: FRAGMENT) : CARDINAL ;
VAR
   s: CARDINAL ;
   f: FRAGMENT ;
BEGIN
   s := 0 ;
   IF h#NIL
   THEN
      f := h ;
      REPEAT
         INC(s) ;
         f := f^.Right
      UNTIL f=h
   END ;
   RETURN( s )
END CountFragments ;


END MetaStore.
