IMPLEMENTATION MODULE Executive ;

(*
    Title      : Executive
    Author     : Gaius Mulley
    System     : UNIX (gm2)
    Date       : Thu Oct  5 12:56:01 1995
    Last edit  : Thu Oct  5 12:56:01 1995
    Description: implement the executive interface by using RTSIM components.

                 SEMAPHORE, DESCRIPTOR,
                 
                 InitProcess, Resume, Suspend, InitSemaphore, Wait, Signal,
                 WaitForIO, Ps, GetCurrentProcess ;
*)

IMPORT M2RTS ;
FROM Debug IMPORT Halt ;
FROM SYSTEM IMPORT ADDRESS, WORD, LISTEN, OnOrOff, TurnInterrupts ;
FROM SysStorage IMPORT ALLOCATE ;


TYPE
   SEMAPHORE  = CARDINAL ;         (* defines dijkstra's semaphores *)
   DESCRIPTOR = POINTER TO WORD ;  (* handle onto a process         *)


(*
   InitProcess - initializes a process which is held in the suspended
                 state. When the process is resumed it will start executing
                 procedure, p. The process has a maximum stack size of,
                 StackSize, bytes and its textual name is, Name.
                 The StackSize should be at least 5000 bytes.
*)

PROCEDURE InitProcess (p: PROC; StackSize: CARDINAL;
                       Name: ARRAY OF CHAR) : DESCRIPTOR ;
VAR
   StackBot: ADDRESS ;
BEGIN
   ALLOCATE(StackBot, StackSize) ;
   ASM
      load.d.d  16(%ra)   #  Name
      load.d.d  12(%ra)   #  StackSize
      load.d.d  -4(%ra)   #  StackBot
      load.d.d  $3        #  priority middle (normal)
      load.d.d   8(%ra)   #  p
      initprocess
      store.d.d %rd       #  return( DESCRIPTOR )
   END
END InitProcess ;


(*
   Resume - resumes a suspended process. If all is successful then the process, p,
            is returned. If it fails then NIL is returned.
*)

PROCEDURE Resume (d: DESCRIPTOR) : DESCRIPTOR ;
BEGIN
   ASM
      load.d.d   8(%ra)   #  d
      resume
      store.d.d  %rd      #  return( DESCRIPTOR )
   END
END Resume ;


(*
   Suspend - suspend the calling process.
             The process can only continue running if another process
             Resumes it.
*)

PROCEDURE Suspend ;
BEGIN
   ASM
      suspend
   END
END Suspend ;


(*
   InitSemaphore - creates a semaphore whose initial value is, v, and
                   whose name is, Name.
*)

PROCEDURE InitSemaphore (v: CARDINAL; Name: ARRAY OF CHAR) : SEMAPHORE ;
BEGIN
   ASM
      load.d.d   8(%ra)   # initial semaphore value
      initsemaphore
      store.d.d %rd       # return( SEMAPHORE )
   END
END InitSemaphore ;


(*
   Wait - performs dijkstra's P operation on a semaphore.
          A process which calls this procedure will
          wait until the value of the semaphore is > 0
          and then it will decrement this value.
*)

PROCEDURE Wait (s: SEMAPHORE) ;
BEGIN
   ASM
      load.d.d  8(%ra)
      wait
   END
END Wait ;


(*
   Signal - performs dijkstra's V operation on a semaphore.
            A process which calls the procedure will increment
            the semaphores value.
*)

PROCEDURE Signal (s: SEMAPHORE) ;
BEGIN
   ASM
      load.d.d  8(%ra)
      signal
   END
END Signal ;


(*
   RotateRunQueue - rotates the process run queue.
                    It does not call the scheduler.
*)

PROCEDURE RotateRunQueue ;
BEGIN
   ASM
      load.d.d  $3      # rotate this middle (normal) queue priority
      rotaterunq
   END
END RotateRunQueue ;


(*
   ProcessName - 
*)

PROCEDURE ProcessName (d: DESCRIPTOR) ;
BEGIN
   (* do nothing - this is not yet implemented (only used for debugging) *)
END ProcessName ;


(*
   DebugProcess - 
*)

PROCEDURE DebugProcess (d: DESCRIPTOR) ;
BEGIN
   (* does nothing yet, should be made to do the same as i[345]86 Executive *)
END DebugProcess ;


(*
   WaitForIO - waits for an interrupt to occur on vector, VectorNo.
*)

PROCEDURE WaitForIO (VectorNo: CARDINAL) ;
BEGIN
   ASM
      # move ourselves onto a higher run q priority
      load.d.d  $7
      putrunq
      load.d.d   8(%ra)   # vectorno (actually port no. in simulator)
      waitforio
   END
END WaitForIO ;


(*
   Ps - displays a process list together with relevant their status.
*)

PROCEDURE Ps ;
BEGIN
   ASM
      monitor
   END
END Ps ;


(*
   GetCurrentProcess - returns the descriptor of the current running
                       process.
*)

PROCEDURE GetCurrentProcess () : DESCRIPTOR ;
BEGIN
   ASM
      getid
      store.d.d   %rd    # RETURN( DESCRIPTOR )
   END
END GetCurrentProcess ;


(*
   IdleProcess - note that we could just use
                 LOOP
                    LISTEN
                 END

                 which will work but it is useful to
                 use the implementation below as it speed up
                 simulation considerably. It also detects deadlock
                 much quicker.
*)

PROCEDURE IdleProcess ;
VAR
   ToOldState: OnOrOff ;
BEGIN
   ToOldState := TurnInterrupts(On) ;
   LOOP
      ASM
         load.d.d   $1000000000
         pr
      END
   END
END IdleProcess ;


(*
   InitIdle - starts the idle process
*)

PROCEDURE InitIdle (p: PROC; StackSize: CARDINAL;
                    Name: ARRAY OF CHAR) ;
VAR
   StackBot: ADDRESS ;
BEGIN
   (* firstly make ourselves run at middle run q priority *)
   ASM
      load.d.d  $3
      putrunq
   END ;
   ALLOCATE(StackBot, StackSize) ;   (* now create a process at runq 0 *)
   ASM
      load.d.d  16(%ra)   #  Name
      load.d.d  12(%ra)   #  StackSize
      load.d.d  -4(%ra)   #  StackBot
      load.d.d  $0        #  priority = lowest
      load.d.d   8(%ra)   #  p
      initprocess
      resume              #  and make it runnable
      store.d.d %rd       #  return( DESCRIPTOR )  [or throw it away]
   END
END InitIdle ;


BEGIN
   (* must start an idle process.. *)
   InitIdle(IdleProcess, 100, 'idle')     (* with a small stack... *)
END Executive.
